
//******************************************************************************
//
//      File             : extensionsForAMRWB.c
//
//      Version          : 5.0.0  
//      Date             : 22 February 2002 
//      Author           : Clemens Suerbaum
//
//      Former Versions  : none
//
//      Description      : source file containing TFO extensions for AMR_WB
//
//******************************************************************************



#include <stdio.h>
#include <stdlib.h>
#include "extensionsForAMRWB.h"
#include "oacs.h"
#include "tfo_decision.h"
#include <iostream.h>                     // for use of cout and cin


// ---------------------------------------------------------------------------------------
//                               functions
// ---------------------------------------------------------------------------------------

// ------------------------------------------------------------------------
//
// Function   compatible_AMRWB_codec_types     new in 5.0.0                        
// Purpose    checks if two codec types are compatible AMR-WB codecs
//
// ------------------------------------------------------------------------
int compatible_AMR_WB_codec_types(CodecType lType, CodecType dType)
	{
	if (
		(lType==FR_AMR_WB   && dType==FR_AMR_WB  )  || 
		(lType==FR_AMR_WB   && dType==UMTS_AMR_WB)  || 
		(lType==UMTS_AMR_WB && dType==FR_AMR_WB  )  ||  
		(lType==UMTS_AMR_WB && dType==UMTS_AMR_WB)  ||
		(lType==FR_AMR_WB   && dType==OFR_AMR_WB)   ||
		(lType==OFR_AMR_WB  && dType==FR_AMR_WB)    ||
		(lType==UMTS_AMR_WB && dType==OFR_AMR_WB)   ||
		(lType==OFR_AMR_WB  && dType==UMTS_AMR_WB)  ||
		(lType==FR_AMR_WB   && dType==OHR_AMR_WB)   ||
		(lType==OHR_AMR_WB  && dType==FR_AMR_WB)    ||
		(lType==UMTS_AMR_WB && dType==OHR_AMR_WB)   ||
		(lType==OHR_AMR_WB  && dType==UMTS_AMR_WB)  ||
		(lType==OFR_AMR_WB  && dType==OHR_AMR_WB)   ||
		(lType==OHR_AMR_WB  && dType==OFR_AMR_WB)   ||
		(lType==OHR_AMR_WB  && dType==OHR_AMR_WB)   ||
		(lType==OFR_AMR_WB  && dType==OFR_AMR_WB)  
		)
		return 1;
	else
		return 0;
	}

// ------------------------------------------------------------------------
//
// Function  wb_mode_isInWBcodecSet
// Purpose   checks if a given WB mode is contained in a WB codec set 
//           (mathematically spoken: checks if a certain power of two is in 
//            in a number in the range of 0..511)
// 
// ------------------------------------------------------------------------

int wb_mode_isInWBcodecSet      (int          wb_m,
								 WB_CodecSet  wb_cs_in)
	{ 
    WB_CodecSet  wb_cs = wb_cs_in;
	
	if (
		(wb_cs < 0)    ||
		(wb_cs > 511)  ||       // WB codec set outside allowed range 0..511
		                        // not allowed value ranges for mode
		(wb_m!=wb_mode_0660 && wb_m!=wb_mode_0885 && wb_m!=wb_mode_1265 &&
		 wb_m!=wb_mode_1425 && wb_m!=wb_mode_1585 && wb_m!=wb_mode_1825 &&
		 wb_m!=wb_mode_1985 && wb_m!=wb_mode_2305 && wb_m!=wb_mode_2385)
		)        
		{
		fprintf(stderr, "ERROR in wb_mode_isInWBcodecSet: wrong input.\n"); 
		exit(0);                
		}

	if (wb_cs_in == 0)
		return 0;         // shortcut for empty codec set

	if (wb_cs == wb_m)             
		return 1;        // shortcut (only this mode in codec set)
	
	
	int k=0;
	int testMode = wb_mode_2385;     // starting with highest mode = 256
			
	for (k=0;k<9;k++)
		{
		if (wb_cs >= testMode)
			{
			wb_cs = wb_cs - testMode;
						
			if (wb_cs == wb_m)             
				return 1;    // mode IS contained in codec set
				
			if ( 
				( (wb_cs < wb_m) && (wb_m == testMode) ) ||
				(wb_cs == wb_m) 
			   )            
				return 1;    // mode IS contained in codec set
				
			if ( (wb_cs < wb_m) && (wb_m < testMode) )             
				return 0;    // mode is NOT contained in codec set

			} // if (wb_cs >= testMode)
 
		if (testMode >= 2)          // just to ensure defined result in last loop
			{
			testMode = testMode / 2 ;   // next lower mode  
			}
		
		}  // end for

	return 0;
	}


// ------------------------------------------------------------------------
//
// Function  noOfWBmodes
// Purpose   count number of modes in a AMR-WB codec set 
// 
// ------------------------------------------------------------------------

int noOfWBmodes(WB_CodecSet wb_codecSet)
	{
	int k=0; 
	int	n=0;
	int testMode = wb_mode_2385;     // starting with highest mode = 256

	for (k=0; k<9; k++)
		{
		if ( wb_mode_isInWBcodecSet(testMode, wb_codecSet) )
			n = n + 1;
		
		if (testMode >= 2)          // just to ensure defined result in last loop
		testMode = testMode / 2 ;   // next lower mode  
		}
	
	return n;
	}


// ------------------------------------------------------------------------
//
// Function   lowest_mode_in_WBcodecSet
// Purpose    determines lowest mode in a WB codec set
//
// Remark     for an empty WB codec set the function returns 0
//
// ------------------------------------------------------------------------
int      lowest_mode_in_WBcodecSet      (WB_CodecSet    wb_cs)
	{
	int k=0; 
	int testMode = 1;     // starting with lowest mode

	for (k=0; k<9; k++)
		{
		if ( wb_mode_isInWBcodecSet(testMode, wb_cs) )
			return testMode;
		else 
			testMode = testMode * 2 ;   // next higher mode  
		}
	
	return 0; // only reached if no mode was contained
	}


// ------------------------------------------------------------------------
//
// Function   wb_contiguous_subset
// Purpose    check if subset is a contiguous subset of superset 
//
// Remark     for an empty sub- or superset the function returns 0
//
// ------------------------------------------------------------------------

int      wb_contiguous_subset           (WB_CodecSet    subset, 
										 WB_CodecSet    superset)
	{
	int k;
	int testMode=1;   // start with lowest mode
	int noOfCommonModes=0;
	int noOfSubsetModes = noOfWBmodes(subset);

	if (subset==0 || superset==0) 
		return 0;

	if (subset==superset)   // a shortcut
		return 1;

	for (k=0;k<8;k++)
		{
		if (
			!wb_mode_isInWBcodecSet(testMode,subset) && 
			wb_mode_isInWBcodecSet(testMode,superset) 
			)
			{
			if ( noOfCommonModes == noOfSubsetModes )
				return 1;  // remaining untested modes are only in superset
			else 
				return 0;  // remaining untested modes not contiguous
			}
		else if (
			wb_mode_isInWBcodecSet(testMode,subset) && 
			wb_mode_isInWBcodecSet(testMode,superset) 
			)
			{
			noOfCommonModes = noOfCommonModes + 1;
			}
		else if (
			wb_mode_isInWBcodecSet(testMode,subset) && 
			!wb_mode_isInWBcodecSet(testMode,superset) 
			)
			{
			return 0;  // superset is no superset
			}
		// else { ; // nothing to do }

		
		testMode = testMode * 2 ;   // next higher mode  


		}  // end for k=0..8
	
	if (noOfCommonModes == noOfSubsetModes)
		return 1;
	else 
		return 0;
}


// ------------------------------------------------------------------------
//
// Function   wb_contiguous_subset_selection
// Purpose    selects a subset out of cs which is contiguous to lacs and dacs
//
// Remark     for an empty subset the function returns 0
//
// ------------------------------------------------------------------------

WB_CodecSet wb_contiguous_subset_selection (
										WB_CodecSet     wb_cs, 
				                        WB_CodecSet     wb_lacs, 
									    WB_CodecSet     wb_dacs, 
									    int             cmacs
									    )
	{
	int testMode=1;   // start with lowest mode;
	int k;
	WB_CodecSet output_CS = 0;
	WB_CodecSet intermediate_CS = 0;

	for (k=0;k<9;k++)
		{
		if ( wb_mode_isInWBcodecSet(testMode,wb_cs) )
			{
			intermediate_CS = output_CS + testMode;

			if	(
				 ( wb_contiguous_subset(intermediate_CS,wb_lacs) ) &&   
				 ( wb_contiguous_subset(intermediate_CS,wb_dacs) ) && 
				 ( noOfWBmodes(intermediate_CS) <= cmacs)
				)
				output_CS = intermediate_CS;

			}

		testMode = testMode * 2 ;   // next higher mode  

		}  // end for
	
	wb_cs = output_CS;
	return wb_cs;
	}


// ------------------------------------------------------------------------
//
// Function   wb_CodecSetIntersection
// Purpose    determines the intersection of two WB codec sets
//
// ------------------------------------------------------------------------
WB_CodecSet wb_CodecSetIntersection     (WB_CodecSet    codecSet1,       
										 WB_CodecSet    codecSet2
										 )    
	{
	WB_CodecSet intersection = 0;   // initialisation
	int testMode = 1;   // start with lowest WB mode
	int k=0;

	if ( (codecSet1 == 0) || (codecSet2 == 0) )
		return intersection;                 // shortcut for empty CS

	if  (codecSet1 == 511)
		return codecSet2;                    // shortcut for full CS1

	if  (codecSet2 == 511)
		return codecSet1;                    // shortcut for full CS2

	if (codecSet1 == codecSet2)
		return codecSet1;                    // shortcut for identical CS
		

	for (k=0; k<9; k++)
		{
		if ((wb_mode_isInWBcodecSet(testMode,codecSet1)==1)&&
			(wb_mode_isInWBcodecSet(testMode,codecSet2)==1)  ) 
			intersection = intersection + testMode;
			
		testMode = testMode * 2;          // moves to next higher mode
		}
	
	return intersection;
	}


// ------------------------------------------------------------------------
//
// Function   wb_acceptable
// Purpose    checks if an AMR-WB ACS is acceptable
//
// ------------------------------------------------------------------------
int wb_acceptable(Data* data, WB_CodecSet wb_cs)
	{
	if	(
			( wb_cs == 0 ) // empty wb_cs
			||
			(	// only 6.65 in wb_cs
				( noOfWBmodes(wb_cs) == 1) && wb_mode_isInWBcodecSet(1, wb_cs) 
			)   
			||
			(	// only mode 15.85 or higher in a GMSK or 8PSK HR channel
				( lowest_mode_in_WBcodecSet(wb_cs) >= 16) && 
				( (data->dType == FR_AMR_WB) || (data->lType == FR_AMR_WB) ||
				  (data->dType == OHR_AMR_WB) || (data->lType == OHR_AMR_WB)  ) 
			)   
		)
		return 0;

	else
		return 1;
	}



// ------------------------------------------------------------------------
//
// Function   wb_tfo_decision_AMR                        5.0.0
// Purpose    decides if AMR-WB TFO is possible and on which ACS it its started
//
// ------------------------------------------------------------------------
int wb_tfo_decision_AMR (Data* data)
	{
	// LACS == DACS and acceptable? 
	if (
		(data->wb_lacs == data->wb_dacs) 
		&& 
		wb_acceptable(data, data->wb_lacs)   
		)
		{
		printf("Immediate TFO on LACS=DACS.\n");
		return 1;
		}

	// no FR - HR - Matching for AMR-WB */

	// IACS == OACS and acceptable? 
	if ( 
		(data->wb_cacs == data->wb_oacs) &&
		wb_acceptable(data, data->wb_cacs)   
		)
		{
		printf("Immediate TFO on IACS=OACS.\n");
		return 1;
		}
	
	// IACS contiguous subset of OACS and acceptable ? 
	if (
		(wb_contiguous_subset(data->wb_cacs, data->wb_oacs)) &&
		wb_acceptable(data, data->wb_cacs)   
		)
		{
		printf("Immediate TFO on IACS, contiguous subset of OACS.\n");
		printf("Do codec mode optimisation when TFO is established \n");
		return 1;
		}

	// OACS acceptable ? 
	if (wb_acceptable(data, data->wb_oacs))
		{
		printf("Change ACS to OACS and establish TFO on OACS.\n");
		return 1;
		}
	else 
		{
		printf("No AMR-WB TFO, because OACS not acceptable.\n");  
		return 0;
		}
	}

// ------------------------------------------------------------------------
//
// Function   weight_sum_of_WBCodecSet                        
// Purpose    calculates the sum of weights for a WB Codec Set
//
// ------------------------------------------------------------------------
int weight_sum_of_WBCodecSet (WB_CodecSet wb_cs, 
							  CodecType   lType,
							  CodecType   dType)
{
	int sum=0;
	int weight_0660= 0;
	int weight_0885= 0;		
	int weight_1265= 0;		
	int weight_1425= 0;		
	int weight_1585= 0;		
	int weight_1825= 0;		
	int weight_1985= 0;	
	int weight_2305= 0;		
	int weight_2385= 0;

	if ( // GMSK or 8PSK-HR is involved
		(lType==FR_AMR_WB)  || 
		(dType==FR_AMR_WB)  ||
		(lType==OHR_AMR_WB) || 
		(dType==OHR_AMR_WB)
		)
		{
		weight_0660= weightFactor_wb_mode_0660_GSM;
		weight_0885= weightFactor_wb_mode_0885_GSM;		
		weight_1265= weightFactor_wb_mode_1265_GSM;		
		weight_1425= weightFactor_wb_mode_1425_GSM;		
		weight_1585= weightFactor_wb_mode_1585_GSM;		
		weight_1825= weightFactor_wb_mode_1825_GSM;		
		weight_1985= weightFactor_wb_mode_1985_GSM;	
		weight_2305= weightFactor_wb_mode_2305_GSM;		
		weight_2385= weightFactor_wb_mode_2385_GSM;
		}					
		
	else 
		{
		weight_0660= weightFactor_wb_mode_0660_UMTS;
		weight_0885= weightFactor_wb_mode_0885_UMTS;		
		weight_1265= weightFactor_wb_mode_1265_UMTS;		
		weight_1425= weightFactor_wb_mode_1425_UMTS;		
		weight_1585= weightFactor_wb_mode_1585_UMTS;		
		weight_1825= weightFactor_wb_mode_1825_UMTS;		
		weight_1985= weightFactor_wb_mode_1985_UMTS;	
		weight_2305= weightFactor_wb_mode_2305_UMTS;		
		weight_2385= weightFactor_wb_mode_2385_UMTS;
		}	

	if (wb_mode_isInWBcodecSet(wb_mode_0660, wb_cs))
		sum = sum + weight_0660;

	if (wb_mode_isInWBcodecSet(wb_mode_0885, wb_cs))
		sum = sum + weight_0885;

	if (wb_mode_isInWBcodecSet(wb_mode_1265, wb_cs))
		sum = sum + weight_1265;
	
	if (wb_mode_isInWBcodecSet(wb_mode_1425, wb_cs))
		sum = sum + weight_1425;
	
	if (wb_mode_isInWBcodecSet(wb_mode_1585, wb_cs))
		sum = sum + weight_1585;
	
	if (wb_mode_isInWBcodecSet(wb_mode_1825, wb_cs))
		sum = sum + weight_1825;
	
	if (wb_mode_isInWBcodecSet(wb_mode_1985, wb_cs))
		sum = sum + weight_1985;
	
	if (wb_mode_isInWBcodecSet(wb_mode_2305, wb_cs))
		sum = sum + weight_2305;
	
	if (wb_mode_isInWBcodecSet(wb_mode_2385, wb_cs))
		sum = sum + weight_2385;

	return sum;
}

// ------------------------------------------------------------------------
//
// Function   default_WB_codec_set                        
// Purpose    calculates default WB Codec Set 
//
// ------------------------------------------------------------------------
WB_CodecSet default_WB_codec_set (
							int noOfModes, 
							CodecType   lType,
							CodecType   dType)
{
	if ( (noOfModes == 0)  || (noOfModes > max_wb_macs) )
		{
		fprintf(stderr,"ERROR: function default_WB_codec_set called with noOfModes=0 or >8\n"); 
		exit(0);   
		}

	int wb_default_ACS=0;

	int weight_0660= 0;
	int weight_0885= 0;		
	int weight_1265= 0;		
	int weight_1425= 0;		
	int weight_1585= 0;		
	int weight_1825= 0;		
	int weight_1985= 0;	
	int weight_2305= 0;		
	int weight_2385= 0;

	int ranking_0660= 0;
	int ranking_0885= 0;		
	int ranking_1265= 0;		
	int ranking_1425= 0;		
	int ranking_1585= 0;		
	int ranking_1825= 0;		
	int ranking_1985= 0;	
	int ranking_2305= 0;		
	int ranking_2385= 0;

	if ( // GMSK or 8PSK-HR is involved
		(lType==FR_AMR_WB)  || 
		(dType==FR_AMR_WB)  ||
		(lType==OHR_AMR_WB) || 
		(dType==OHR_AMR_WB)
		)
		{
		weight_0660= weightFactor_wb_mode_0660_GSM;
		weight_0885= weightFactor_wb_mode_0885_GSM;		
		weight_1265= weightFactor_wb_mode_1265_GSM;		
		weight_1425= weightFactor_wb_mode_1425_GSM;		
		weight_1585= weightFactor_wb_mode_1585_GSM;		
		weight_1825= weightFactor_wb_mode_1825_GSM;		
		weight_1985= weightFactor_wb_mode_1985_GSM;	
		weight_2305= weightFactor_wb_mode_2305_GSM;		
		weight_2385= weightFactor_wb_mode_2385_GSM;
		}					
		
	else 
		{
		weight_0660= weightFactor_wb_mode_0660_UMTS;
		weight_0885= weightFactor_wb_mode_0885_UMTS;		
		weight_1265= weightFactor_wb_mode_1265_UMTS;		
		weight_1425= weightFactor_wb_mode_1425_UMTS;		
		weight_1585= weightFactor_wb_mode_1585_UMTS;		
		weight_1825= weightFactor_wb_mode_1825_UMTS;		
		weight_1985= weightFactor_wb_mode_1985_UMTS;	
		weight_2305= weightFactor_wb_mode_2305_UMTS;		
		weight_2385= weightFactor_wb_mode_2385_UMTS;
		}	

	//--------------------------------------------------------
	// Determine ranking of modes 
	//--------------------------------------------------------	
	ranking_0660 =	(weight_0660 > weight_0660) +
					(weight_0660 > weight_0885) +
					(weight_0660 > weight_1265) +
					(weight_0660 > weight_1425) +
					(weight_0660 > weight_1585) +
					(weight_0660 > weight_1825) +
					(weight_0660 > weight_1985) +
					(weight_0660 > weight_2305) +
					(weight_0660 > weight_2385) 
					;
	if (ranking_0660 > (8 - noOfModes) )
		{ 
		wb_default_ACS = wb_default_ACS + wb_mode_0660;
		}

	ranking_0885 =	(weight_0885 > weight_0660) +
					(weight_0885 > weight_0885) +
					(weight_0885 > weight_1265) +
					(weight_0885 > weight_1425) +
					(weight_0885 > weight_1585) +
					(weight_0885 > weight_1825) +
					(weight_0885 > weight_1985) +
					(weight_0885 > weight_2305) +
					(weight_0885 > weight_2385) 
					;
	if (ranking_0885 > (8 - noOfModes) )
		{ 
		wb_default_ACS = wb_default_ACS + wb_mode_0885;
		}

	ranking_1265 =	(weight_1265 > weight_0660) +
					(weight_1265 > weight_0885) +
					(weight_1265 > weight_1265) +
					(weight_1265 > weight_1425) +
					(weight_1265 > weight_1585) +
					(weight_1265 > weight_1825) +
					(weight_1265 > weight_1985) +
					(weight_1265 > weight_2305) +
					(weight_1265 > weight_2385) 
					;
	if (ranking_1265 > (8 - noOfModes) )
		{ 
		wb_default_ACS = wb_default_ACS + wb_mode_1265;
		}
	
	ranking_1425 =	(weight_1425 > weight_0660) +
					(weight_1425 > weight_0885) +
					(weight_1425 > weight_1265) +
					(weight_1425 > weight_1425) +
					(weight_1425 > weight_1585) +
					(weight_1425 > weight_1825) +
					(weight_1425 > weight_1985) +
					(weight_1425 > weight_2305) +
					(weight_1425 > weight_2385) 
					;
	if (ranking_1425 > (8 - noOfModes) )
		{ 
		wb_default_ACS = wb_default_ACS + wb_mode_1425;
		}
	
	ranking_1585 =	(weight_1585 > weight_0660) +
					(weight_1585 > weight_0885) +
					(weight_1585 > weight_1265) +
					(weight_1585 > weight_1425) +
					(weight_1585 > weight_1585) +
					(weight_1585 > weight_1825) +
					(weight_1585 > weight_1985) +
					(weight_1585 > weight_2305) +
					(weight_1585 > weight_2385) 
					;
	if (ranking_1585 > (8 - noOfModes) )
		{ 
		wb_default_ACS = wb_default_ACS + wb_mode_1585;
		}

	ranking_1825 =	(weight_1825 > weight_0660) +
					(weight_1825 > weight_0885) +
					(weight_1825 > weight_1265) +
					(weight_1825 > weight_1425) +
					(weight_1825 > weight_1585) +
					(weight_1825 > weight_1825) +
					(weight_1825 > weight_1985) +
					(weight_1825 > weight_2305) +
					(weight_1825 > weight_2385) 
					;
	if (ranking_1825 > (8 - noOfModes) )
		{ 
		wb_default_ACS = wb_default_ACS + wb_mode_1825;
		}

	ranking_1985 =	(weight_1985 > weight_0660) +
					(weight_1985 > weight_0885) +
					(weight_1985 > weight_1265) +
					(weight_1985 > weight_1425) +
					(weight_1985 > weight_1585) +
					(weight_1985 > weight_1825) +
					(weight_1985 > weight_1985) +
					(weight_1985 > weight_2305) +
					(weight_1985 > weight_2385) 
					;
	if (ranking_1985 > (8 - noOfModes) )
		{ 
		wb_default_ACS = wb_default_ACS + wb_mode_1985;
		}

	ranking_2305 =	(weight_2305 > weight_0660) +
					(weight_2305 > weight_0885) +
					(weight_2305 > weight_1265) +
					(weight_2305 > weight_1425) +
					(weight_2305 > weight_1585) +
					(weight_2305 > weight_1825) +
					(weight_2305 > weight_1985) +
					(weight_2305 > weight_2305) +
					(weight_2305 > weight_2385) 
					;
	if (ranking_2305 > (8 - noOfModes) )
		{ 
		wb_default_ACS = wb_default_ACS + wb_mode_2305;
		}

	ranking_2385 =	(weight_2385 > weight_0660) +
					(weight_2385 > weight_0885) +
					(weight_2385 > weight_1265) +
					(weight_2385 > weight_1425) +
					(weight_2385 > weight_1585) +
					(weight_2385 > weight_1825) +
					(weight_2385 > weight_1985) +
					(weight_2385 > weight_2305) +
					(weight_2385 > weight_2385) 
					;
	if (ranking_2385 > (8 - noOfModes) )
		{ 
		wb_default_ACS = wb_default_ACS + wb_mode_2385;
		}
	
	return wb_default_ACS;
}